#include "FragmentDownloadThread.hpp"
#include "OppStoreUtils.hpp"
#include "Benchmark.hpp"
#include "utils/c++/Config.hpp"
#include "utils/c++/NoSuchConfigException.hpp"

#include <openssl/sha.h>
#include <cassert>

#include <set>
#include <iostream>
#include <string>
using namespace std;

/**
 * Used only in the experiments
 */
set<string> allowedAdrsSet;

map<int, FragmentDownloadThread *> downloadThreadMap;
int nextDownloadThreadId = 1;

AdrDataTransferStub *adrDataTransferStub;
DataRetrievalManager *dataRetrievalManager;

FragmentDownloadThread::FragmentDownloadThread
( string ipAddress_, short portNumber_, int fragmentNumber_, CodedDataInfo *dataInfo_, int fragmentPos_ )
{
    
    this->dataInfo = dataInfo_;
    this->ipAddress = ipAddress_;
    this->portNumber = portNumber_;
    this->hashKey = dataInfo_->fragmentKeyList[fragmentNumber_];
    this->fragmentPos = fragmentPos_;
    this->fragmentNumber = fragmentNumber_;
    this->fragmentSize = dataInfo_->fragmentSizeList[fragmentNumber_];
    this->remainingDownloads = dataInfo_->remainingDownloads;
}

FragmentDownloadThread::~FragmentDownloadThread()
{
}

void FragmentDownloadThread::configureDownloadThreads( AdrDataTransferStub *adrDataTransferStub_, DataRetrievalManager *dataRetrievalManager_ ) {
    adrDataTransferStub  = adrDataTransferStub_;
    dataRetrievalManager = dataRetrievalManager_;
}

void FragmentDownloadThread::performDownload() {

    Benchmark::getCurrentBenchmark()->fragmentStart[fragmentPos] = OppStoreUtils::timeInMillis();
    
    char *hexKey = new char[ OppStoreUtils::binaryKeySize*2  ];
    OppStoreUtils::convertBinaryToHex(hashKey, hexKey);

    cout << "Downloading fragment from " << ipAddress << ":" << portNumber << " key=";
    OppStoreUtils::printHexKey(hexKey, cout);
    cout << " fragmentSize=" << fragmentSize << "." << endl;

    //void * fragmentData = new char[fragmentSize];
    //dataInfo->encData[fragmentPos] = (char *)fragmentData;
    void * fragmentData = dataInfo->encData[fragmentPos];
    int tmpSize = adrDataTransferStub->readData(ipAddress, portNumber, hexKey, OppStoreUtils::binaryKeySize*2, fragmentData, fragmentSize, dataInfo->availableBytes[fragmentPos]);
    delete[] hexKey;

    /**
     * Checks if fragment size and hash are correct
     */
    assert (fragmentSize == tmpSize);
        
    char *fragmentKey = new char[OppStoreUtils::binaryKeySize];
    SHA1( (unsigned char *)fragmentData, fragmentSize, (unsigned char *)fragmentKey);     
       
    for (int keyPos=0; keyPos<OppStoreUtils::binaryKeySize; keyPos++)
        if ( hashKey[keyPos] != fragmentKey[keyPos] ) {
            cerr << "ERROR: checksum for fragment does not match at position " << keyPos << endl;
            cerr << "with fragmentNumber " << fragmentNumber << " at position " << fragmentPos << " ipAddress=" << ipAddress << ":" << portNumber << "." << endl;
            exit(-1);
        }
        
    delete[] fragmentKey;
    delete[] hashKey;
                                
    /**
     * Puts the checked data into the dataInfo for later decoding
     */  
    //dataInfo->recoveredFragmentIndexes[fragmentPos] = fragmentNumber;
    cout << " Added fragment Number " << fragmentNumber << " at position " << fragmentPos << "." << endl;
   
    Benchmark::getCurrentBenchmark()->fragmentFinish[fragmentPos] = OppStoreUtils::timeInMillis();
     
    (*remainingDownloads)--;
    cout << "remainingDownloads=" << *remainingDownloads << endl;
   
}

void *FragmentDownloadThread::run( void *ptr ) {
 
    int *threadId = (int *)ptr;
    FragmentDownloadThread *downloadThread = downloadThreadMap[*threadId];
    downloadThread->performDownload();
        
    downloadThreadMap.erase(*threadId);
    delete threadId;

    return NULL;
}


// 143.107.1.1:9003
int FragmentDownloadThread::setAllowedAdrs(vector<string> & allowedAdrs) {
    
    allowedAdrsSet.clear();
    for (uint fragment=0; fragment < allowedAdrs.size(); fragment++) {
        allowedAdrsSet.insert( allowedAdrs[fragment] );
    }
    
    return 0;
}

void *FragmentDownloadThread::launchDownloadThread( void *ptr ) {

//    for (unsigned int i=0; i<adrAddresses.size(); i++)
//        cout << adrAddresses[i] << " ";
//    cout << endl;
    int requestId = *( int *)ptr;
    delete (int *)ptr;
    CodedDataInfo *dataInfo = dataRetrievalManager->getCodedDataInfo(requestId );    

    dataInfo->remainingDownloads = new int;
    *(dataInfo->remainingDownloads) = dataInfo->neededFragments; // adrAddresses.size();
    dataInfo->encData.resize(dataInfo->neededFragments);
    dataInfo->fragmentSizeList.resize(dataInfo->neededFragments);
    dataInfo->recoveredFragmentIndexes.resize(dataInfo->neededFragments);
    dataInfo->availableBytes.resize(dataInfo->neededFragments);    
    
    //cout << "neededFragments: " << *(dataInfo->remainingDownloads) << endl; 

    Benchmark *benchmark = Benchmark::getCurrentBenchmark();
    benchmark->nFragmentTimes = dataInfo->neededFragments;    
    benchmark->fragmentStart  = new long[dataInfo->neededFragments];
    benchmark->fragmentFinish = new long[dataInfo->neededFragments];
    benchmark->fragmentAddress.resize(dataInfo->neededFragments);

    // Poderia ter uma lista de servidores preferenciais de download (ip+port)
    // Neste caso, o broker baixaria dados apenas destes servidores

    // file allowedAdrs.conf:
    // restrictAdrs false
    // 123.45.67.87 9001     
                    
    int nRecovered = 0;                    
    for (uint fragment=0; fragment < dataInfo->adrAddresses.size() && nRecovered < dataInfo->neededFragments; fragment++) {
    //for (int fragment=0; fragment < dataInfo->neededFragments; fragment++) {

        if ( allowedAdrsSet.size() > 0 && allowedAdrsSet.find(dataInfo->adrAddresses[fragment]) == allowedAdrsSet.end() ) {
            cout << "Discarding repository " << dataInfo->adrAddresses[fragment] << "." << endl; 
            continue;
        }

        cout << "Downloading from repository " << dataInfo->adrAddresses[fragment] << " fragment=" << fragment << "." << endl;

        string ipAddress;
        short portNumber;
        OppStoreUtils::extractIpAddress(dataInfo->adrAddresses[fragment], ipAddress, portNumber);

        dataInfo->encData[nRecovered] = new char[ dataInfo->fragmentSizeList[fragment] ];     
        dataInfo->recoveredFragmentIndexes[nRecovered] = fragment;           
        dataInfo->availableBytes[nRecovered] = new long;
        *(dataInfo->availableBytes[nRecovered]) = 0; 
             
        Benchmark::getCurrentBenchmark()->fragmentAddress[nRecovered] = dataInfo->adrAddresses[fragment];
                                        
        /**
         * Launches the fragment download thread.
         */ 
        FragmentDownloadThread *downloadThread = 
            new FragmentDownloadThread(ipAddress, portNumber, fragment, dataInfo, nRecovered );
        int *threadId = new int;
        *threadId = nextDownloadThreadId++;
        downloadThreadMap[*threadId] = downloadThread;
             
        pthread_t thread1;
        pthread_create( &thread1, NULL, FragmentDownloadThread::run, threadId );
        pthread_detach(thread1);
        
        nRecovered++;
        //sleep(1);    
    }        
    
    /**
     * Waits for the downloads to finish
     */   
    //OppStoreUtils::waitVar(100, dataInfo->remainingDownloads);
    
    //cout << "Data retrieval request " << requestId << " completed." << endl;
    
    if (dataInfo->operationCallback != NULL)
        dataInfo->operationCallback->finishedDataOperation(dataInfo, 0);
        
    return NULL;        
}
