#include "RetrievalRequestIda.hpp"
#include "OppStoreUtils.hpp"
#include "Benchmark.hpp"
#include <openssl/sha.h>
#include <cassert>

RetrievalRequestIda::RetrievalRequestIda(int neededFragments, int totalFragments, DataRetrievalManager *dataRetrievalManager) {
    this->dataRetrievalManager_ = dataRetrievalManager;
    this->neededFragments_      = neededFragments;
    this->totalFragments_       = totalFragments;
    this->idaImpl               = IDAImpl::getInstance();
    this->dataInfoFreed         = false;
}

RetrievalRequestIda::~RetrievalRequestIda() {
}

void RetrievalRequestIda::retrieveData (const char *key, void * & data, long & dataSize, void(*appCallback)(int), bool wait) {
  
    // Currently the broker only supports blocking retrieval calls
    assert (wait == true);
  
    cout << "RetrievalRequestManager::retrieveData" << endl;
      
    CodedDataInfo *dataInfo = new CodedDataInfo();
    dataInfo->appCallback = appCallback;
    dataInfo->operationCallback = this;
    if (wait == true) {
        dataInfo->callerMutex = new pthread_mutex_t;
        pthread_mutex_init (dataInfo->callerMutex, NULL);
        dataInfo->callerCond  = new pthread_cond_t;
        pthread_cond_init (dataInfo->callerCond, NULL);
    }
    dataInfo->fileKey = new char[OppStoreUtils::binaryKeySize];
    OppStoreUtils::convertHexToBinary(key, (char *)dataInfo->fileKey);
    //SHA1( (unsigned char*)id, strlen(id), (unsigned char*)dataInfo->fileKey);

    Benchmark::getCurrentBenchmark()->nFragments = totalFragments_;
    Benchmark::getCurrentBenchmark()->nRecover   = neededFragments_;
                
    /**
     * Request data retrieval.
     */ 
    if (wait == true)                
        pthread_mutex_lock( dataInfo->callerMutex );
    
    dataInfo->neededFragments = this->neededFragments_;
    Benchmark::getCurrentBenchmark()->locationStart = OppStoreUtils::timeInMillis();
    this->dataRetrievalManager_->retrieveData( dataInfo );
    Benchmark::getCurrentBenchmark()->locationFinish = OppStoreUtils::timeInMillis();
    
    if (wait == true) {
        pthread_cond_wait( dataInfo->callerCond, dataInfo->callerMutex );
        pthread_mutex_unlock( dataInfo->callerMutex );
       
        cout << "Thread released!" << endl << flush;
           
        /**
         * Tests the recovered data
         */ 
        char *recoveredKey = new char[OppStoreUtils::binaryKeySize];
        SHA1( (unsigned char*)dataInfo->data, dataInfo->dataSize, (unsigned char*)recoveredKey);
        
        //cout << "Testing recovered file key for size=" << dataInfo->dataSize << "." << endl << flush;
        
        for (int i=0; i<OppStoreUtils::binaryKeySize; i++)
            if( ((char *)recoveredKey)[i] != ((char *)dataInfo->fileKey)[i] ) {
                cerr << "ERROR: checksum for file does not match at position " << i << endl;
                exit(-1);
            }

       
       //cout << "File key is correct!" << endl << flush;
               
       data = dataInfo->data;
       dataSize = dataInfo->dataSize;
       releaseMemory(dataInfo);
    }    
    
    cout << "RetrievalRequestIda::retrieveData -> Finished recovering data!" << endl << flush;
}

void RetrievalRequestIda::releaseMemory( CodedDataInfo *dataInfo ) {
    
    this->dataInfoFreed = true;
    
    delete[] dataInfo->fileKey;
    if (dataInfo->callerMutex != NULL) delete dataInfo->callerMutex;
    if (dataInfo->callerCond  != NULL) delete dataInfo->callerCond;
    delete dataInfo;
    
}
    
void RetrievalRequestIda::finishedDataOperation( CodedDataInfo *dataInfo, int status ) {
            
    /**
     * Process the recovered data.
     */
    unsigned char ** recoveredData = new unsigned char *[neededFragments_];
    int *fragmentNumbers = new int[dataInfo->neededFragments];
    for (int i=0; i<neededFragments_; i++) {
        recoveredData[i] = (unsigned char *)dataInfo->encData[i];
        fragmentNumbers[i] = dataInfo->recoveredFragmentIndexes[i];
    }   
    
    cout << "Decoding data!" << endl;     
   
    Benchmark::getCurrentBenchmark()->idaStart  = OppStoreUtils::timeInMillis();             
    dataInfo->data = idaImpl->decodeDataInc(recoveredData, dataInfo->dataSize, fragmentNumbers, neededFragments_, totalFragments_-neededFragments_, dataInfo->availableBytes);
    Benchmark::getCurrentBenchmark()->idaFinish = OppStoreUtils::timeInMillis();

    OppStoreUtils::waitVar(100, dataInfo->remainingDownloads);

    for (int i=0; i<neededFragments_; i++)
        delete[] (dataInfo->encData[i]);
    delete[] fragmentNumbers;
    delete[] recoveredData;
    
    cout << "Releasing calling thread!" << endl ;
        
    /**
     * Releases the waiting thread.
     */
    void (*appCallback) (int) = dataInfo->appCallback;
    if (dataInfo->callerMutex != NULL && dataInfo->callerCond != NULL) {        
        pthread_mutex_lock( dataInfo->callerMutex );
          pthread_cond_signal( dataInfo->callerCond );                 
        pthread_mutex_unlock( dataInfo->callerMutex );       
    }
    else {
        releaseMemory(dataInfo);
    }

    //cout << "Calling back application" << endl << flush;
    
    if (appCallback != NULL)
        appCallback(status);

    //cout << "OK" << endl << flush;
    
}

