#include "StorageRequestIda.hpp"
#include "OppStoreUtils.hpp"
#include "Benchmark.hpp"
#include <openssl/sha.h>
#include <cstring>
 
StorageRequestIda::StorageRequestIda(int neededFragments, int totalFragments, DataStorageManager *dataStorageManager) {
    
    this->dataStorageManager_ = dataStorageManager;
    this->neededFragments_    = neededFragments;
    this->totalFragments_     = totalFragments;
    this->idaImpl             = IDAImpl::getInstance();
}

StorageRequestIda::~StorageRequestIda() {}

void StorageRequestIda::storeData (char * & key, void *data, long dataSize, void(*appCallback)(int), bool wait) {

    cout << "StorageRequestManager::storeData" << endl;
    
    CodedDataInfo *dataInfo = new CodedDataInfo();
    dataInfo->appCallback = appCallback;
    dataInfo->operationCallback = this;
    if (wait == true) {
        dataInfo->callerMutex = new pthread_mutex_t;
        pthread_mutex_init (dataInfo->callerMutex, NULL);
        dataInfo->callerCond  = new pthread_cond_t;
        pthread_cond_init (dataInfo->callerCond, NULL);
    }
    dataInfo->fileKey = new char[OppStoreUtils::binaryKeySize];
    SHA1( (unsigned char*)data, dataSize, (unsigned char*)dataInfo->fileKey);
    dataInfo->dataSize = dataSize;

    key = new char[OppStoreUtils::binaryKeySize*2+1]; 
    OppStoreUtils::convertBinaryToHex(dataInfo->fileKey, key);
    key[OppStoreUtils::binaryKeySize*2] = 0;

    Benchmark::getCurrentBenchmark()->nFragments = totalFragments_;
    Benchmark::getCurrentBenchmark()->nRecover   = neededFragments_;
    /**
     * Encodes the provided data using IDA.
     */
    dataInfo->remainingCodings = new int;
    encodingThread = IDAEncodingThread::launchEncodingThread(dataInfo->remainingCodings, (unsigned char *)data, dataSize, neededFragments_, totalFragments_-neededFragments_);
    int fragmentSize = (dataSize%neededFragments_ == 0) ? dataSize/neededFragments_ : dataSize/neededFragments_+1;
    
    Benchmark::getCurrentBenchmark()->fragmentSize = fragmentSize;
    
    /**
     * Complete the dataInfo fields.
     */
    unsigned char *fragmentShaSource = new unsigned char[OppStoreUtils::binaryKeySize+1];
    for (int i=0; i<OppStoreUtils::binaryKeySize; i++)
        fragmentShaSource[i] = dataInfo->fileKey[i]; 
                 
    for (int i=0; i<totalFragments_; i++) {
        
        unsigned char *fragmentData;
        long *bytesRead;
        encodingThread->getFragmentData(i, fragmentData, bytesRead);        
        
        dataInfo->encData.push_back( (char *)fragmentData );
        dataInfo->availableBytes.push_back( bytesRead );
        dataInfo->fragmentSizeList.push_back( fragmentSize );

        fragmentShaSource[OppStoreUtils::binaryKeySize] = i;                        
        unsigned char *fragmentKey = new unsigned char[OppStoreUtils::binaryKeySize]; 
        SHA1( fragmentShaSource, OppStoreUtils::binaryKeySize+1, fragmentKey );
        dataInfo->fragmentKeyList.push_back( (char*)fragmentKey );
        
        /**
         * Distributes the fragments ids homogeneously over the id space.
         * Used in experiments to force fragments into different and deterministic clusters.
         */
         //fragmentKey[0] = 256/totalFragments_ * i;
         //fragmentKey[OppStoreUtils::binaryKeySize-1] = 256/totalFragments_ * i;
    }

    /**
     * Waits the coding to finish
     */    
    //OppStoreUtils::waitVar(100, dataInfo->remainingCodings);

    if (wait == true)                
        pthread_mutex_lock( dataInfo->callerMutex );
    
    Benchmark::getCurrentBenchmark()->locationStart = OppStoreUtils::timeInMillis();
    this->dataStorageManager_->storeData(dataInfo);
    Benchmark::getCurrentBenchmark()->locationFinish = OppStoreUtils::timeInMillis();
    
    if (wait == true) {
       pthread_cond_wait( dataInfo->callerCond, dataInfo->callerMutex );
       pthread_mutex_unlock( dataInfo->callerMutex );
       releaseMemory(dataInfo);
    }
    else
        OppStoreUtils::waitVar(100, dataInfo->remainingCodings);
}

void StorageRequestIda::releaseMemory( CodedDataInfo *dataInfo ) {

    for (int i=0; i<totalFragments_; i++) {
        delete[] dataInfo->encData[i];
        delete[] dataInfo->fragmentKeyList[i];
    }
    delete[] dataInfo->fileKey;
    delete dataInfo->remainingUploads;
    delete dataInfo->remainingCodings;
    if (dataInfo->callerMutex != NULL) delete dataInfo->callerMutex;
    if (dataInfo->callerCond  != NULL) delete dataInfo->callerCond;
    delete dataInfo;
    
    delete encodingThread;
}

void StorageRequestIda::finishedDataOperation( CodedDataInfo *dataInfo, int status ) {

    void (*appCallback) (int) = dataInfo->appCallback;
    if (dataInfo->callerMutex != NULL && dataInfo->callerCond != NULL) {
            
        pthread_mutex_lock( dataInfo->callerMutex );
          pthread_cond_signal( dataInfo->callerCond );
       pthread_mutex_unlock( dataInfo->callerMutex );
    }
    else {
        releaseMemory(dataInfo);
    }
    if (appCallback != NULL)
        appCallback(status);
}
