package br.usp.ime.oppstore.cdrm;

import java.util.HashMap;
import java.util.Vector;

import org.apache.log4j.Logger;

import br.usp.ime.oppstore.FileInformation;
import br.usp.ime.oppstore.FileFragmentIndex;
import br.usp.ime.oppstore.InvalidFragmentKeyException;
import br.usp.ime.oppstore.cdrm.CdrmApp.OpStoreFragmentProtocol;
import br.usp.ime.oppstore.cdrm.CdrmApp.OpStoreIdProtocol;
import br.usp.ime.oppstore.corba.AccessBroker;
import br.usp.ime.oppstore.message.FileFragmentIndexMessage;
import br.usp.ime.oppstore.message.StoreFragmentListMessage;
import br.usp.ime.oppstore.message.StoreFragmentMessage;
import br.usp.ime.virtualId.VirtualNode;

import rice.p2p.commonapi.Id;

/**
 * Allows the creation and management of file storage in the network.
 * Used when creating FFIs, and sending the FFI and fragments for storage in the network.
 * 
 * @author rcamargo
 *
 */
public class FileStorageRetrievalManager {
    
    private RequestSourceBrokerManager requestSourceBrokerManager;
    private HashMap <Integer, FileInformation> fileInformationMap;
    private VirtualNode virtualNode;
    private Logger logger;
    
    public FileStorageRetrievalManager ( VirtualNode virtualNode ) {
    	this.virtualNode = virtualNode;
        this.fileInformationMap = new HashMap <Integer, FileInformation>();
        this.requestSourceBrokerManager = new RequestSourceBrokerManager();
        this.logger = Logger.getLogger("request." + virtualNode.getNode().getId().toString().substring(0, 9) + ">");
    }
    
    public void storeFileInformationIndex(int requestNumber, int[] notStoredFragmentIndexList, byte[][] fragmentHashList) {
        
        FileInformation fileInformation = fileInformationMap.get(requestNumber);
        
        // TODO: Implement what happens when some fragments are not stored
        FileFragmentIndex fileFragmentIndex = fileInformation.createFileFragmentIndex();
        fileFragmentIndex.updateFragmentHashes(fragmentHashList);
        
        if (notStoredFragmentIndexList.length > 0)
        	fileFragmentIndex.removeFragment(notStoredFragmentIndexList);        

        // Calls the CDRM method to store the fileInformation
        FileFragmentIndexMessage ffiMessage = new FileFragmentIndexMessage(
                fileInformation.fileId, fileFragmentIndex, true, FileInformation.numberOfReplicas,
                virtualNode.getNode().getLocalNodeHandle(), requestNumber); 

        logger.debug("FileStorageRetrievalManager -> Storing FileFragmentIndex for file " + fileInformation.fileId + ".");
        
        virtualNode.routeMessage(fileInformation.fileId, ffiMessage);
    }

    public int getRequestNumber(AccessBroker source) {
    	return requestSourceBrokerManager.getRequestNumber(source);
    }
    
	public int requestFragmentStorageLocations(
			int requestNumber, int fileSize, int[] fragmentSizeArray, Id fileId, Vector<Id> fragmentIdList, int timeoutMinutes, boolean storeGlobal) {
		        
        fileInformationMap.put(new Integer(requestNumber), new FileInformation(fileId, fragmentIdList, fileSize, fragmentSizeArray));

        StringBuffer fragmentList = new StringBuffer();
        for ( Id id : fragmentIdList) {
        	fragmentList.append(id);
        	fragmentList.append(" ");
        }
        logger.debug("FileStorageRetrievalManager -> Requesting fragment storage locations for file " + fileId + ".\n" + fragmentList);
        
        if ( CdrmApp.fragmentProtocol == OpStoreFragmentProtocol.INDEPENDENT ) {

        	for (int i=0; i < fragmentIdList.size(); i++) {       
                StoreFragmentMessage storeMessage = new StoreFragmentMessage (
                        fileId, fragmentIdList.get(i), fragmentSizeArray[i], virtualNode.getNode().getLocalNodeHandle(), requestNumber, timeoutMinutes);
                
                if (CdrmApp.idProtocol == OpStoreIdProtocol.PASTRY)
                	storeMessage.routingVirtualSpaceNumber = 0;
                	
                if (storeGlobal)
                	virtualNode.routeMessage(fragmentIdList.get(i), storeMessage);
                else
                	virtualNode.sendDirectMessage(virtualNode.getNode().getLocalHandle(), storeMessage);
            }
        }
        
        else if ( CdrmApp.fragmentProtocol == OpStoreFragmentProtocol.NEIGHBORHOOD ) {
            Vector<Long> fragmentSizeList = new Vector<Long>();
            for (long fragmentSize : fragmentSizeArray)
                fragmentSizeList.add(fragmentSize);
            
            StoreFragmentListMessage storeFragmentListMessage = new StoreFragmentListMessage (
                    fileId, fragmentIdList, fragmentSizeList, virtualNode.getNode().getLocalNodeHandle(), requestNumber);

            if (storeGlobal)
            	virtualNode.routeMessage(fileId, storeFragmentListMessage);            
            else
            	virtualNode.sendDirectMessage(virtualNode.getNode().getLocalHandle(), storeFragmentListMessage);

        }
		return requestNumber;
	}
	
    public int retrieveFileInformationIndex(Id fileId, AccessBroker sourceAccessBroker) {
        
        int requestNumber = requestSourceBrokerManager.getRequestNumber(sourceAccessBroker);        

        FileFragmentIndexMessage fileInformationMessage = new FileFragmentIndexMessage (
                fileId, null, false, FileInformation.numberOfReplicas, virtualNode.getNode().getLocalNodeHandle(), requestNumber);
        
        logger.debug("FileStorageRetrievalManager -> Retrieving FileFragmentIndex for file " + fileId + ".");
        
        // routes the message in the Network                
        virtualNode.routeMessage(fileId, fileInformationMessage);
        
        return requestNumber;
    }


    public void setFileIndexResponseReceived( FileFragmentIndexMessage message ) {
    	
    	if ( message.isStoreRequest ) {

    		AccessBroker sourceBroker = 
    			requestSourceBrokerManager.getRequestSource(message.messageNumber);

    		FileInformation fileInformation = fileInformationMap.get(message.messageNumber);                
    		fileInformation.numberOfStoredReplicas++;

    		logger.debug("FileStorageRetrievalManager -> Received FileInformationIndex storage confirmation for file " + fileInformation.fileId 
    				+ " (" + fileInformation.numberOfStoredReplicas + " out of " + FileInformation.requiredReplicas + ")");
    		
    		/**
    		 * TODO: Need to treat the case where some servers do not respond
    		 */
    		if (fileInformation.numberOfStoredReplicas == FileInformation.requiredReplicas)
    			sourceBroker.setFileStorageRequestCompleted(message.messageNumber);                   
    		
    		if (fileInformation.numberOfStoredReplicas == FileInformation.numberOfReplicas)
    			fileInformationMap.remove(message.messageNumber);
    	}

    	else {
            AccessBroker sourceBroker = 
                requestSourceBrokerManager.getRequestSource(message.messageNumber);
            
            if ( message.fileFragmentIndex != null ) {            	
        		logger.debug("FileStorageRetrievalManager -> Successfully retrieved FileInformationIndex for file " + message.fileId + ".");      
        		FileFragmentIndex ffi = message.fileFragmentIndex;
                sourceBroker.downloadFragments( message.messageNumber, ffi.adrAddressList, ffi.fragmentKeyList, ffi.fileSize, ffi.fragmentSizeList );
            }
            else {
        		logger.debug("FileStorageRetrievalManager -> Failed to retrieve FileInformationIndex for file " + message.fileId + ".");
                sourceBroker.setFileRetrievalRequestFailed(message.messageNumber);
            }

    	}
    		

    }
        
    public void addFragmentStorageLocation(int requestNumber, Id fragmentId, String adrAddress) {

        FileInformation fileInformation = fileInformationMap.get(new Integer(requestNumber));        
        if (fileInformation == null) return;
    
        try { fileInformation.setFragmentStorageLocation(fragmentId, adrAddress); } 
        catch (InvalidFragmentKeyException e) { e.printStackTrace(); }

        logger.debug("FileStorageRetrievalManager -> Received fragment storage location for file " + fileInformation.fileId 
        		+ " (" + fileInformation.nStorageAddresses + " out of " + fileInformation.adrAddressMap.size() + ")");
        
        if (fileInformation.hasAllStorageLocations() == true) {

        	logger.info("FileStorageRetrievalManager -> Received storage location of all fragments for file " + fileInformation.fileId + ".");
        	
            FileFragmentIndex fileFragmentIndex = fileInformation.createFileFragmentIndex();

            AccessBroker sourceAccessBroker = requestSourceBrokerManager.getRequestSource(requestNumber);            
            assert(sourceAccessBroker != null);
            assert(fileFragmentIndex != null);
            assert(fileFragmentIndex.adrAddressList != null);
            
            sourceAccessBroker.uploadFragments( requestNumber, fileFragmentIndex.adrAddressList );            
        }
        
    }       
}
