package br.usp.ime.oppstore.simulation;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.util.Vector;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;

import br.usp.ime.oppstore.simulation.broker.AccessBrokerSimulatorRemote;

public class OppStoreSimulator implements OppStoreSimulatorRemote {

    /**
     * Contains the list of accessBrokers to which the simulator can send storage and 
     * retrieval requests
     */
    private Vector<AccessBrokerSimulatorRemote> accessBrokerList;
    
    private InetSocketAddress bootstrapNodeAddress;
    
    private AtomicInteger nRegisteredMachines; 
    private AtomicInteger nStorageResponses;
    private AtomicInteger nRetrievalResponses;
    
    private AtomicBoolean isSimulationReadyForStarting;
    private AtomicBoolean isSimulationFinished;
    
    private Logger logger;
    
    private int nRemoteMachines = 1;
    
    private int nFiles;
    private int nFragments;
    private int nNeededFragments;

    public OppStoreSimulator (AtomicBoolean isSimulationFinished, int nStorages, int nFragments, int nNeeded) {
        
    	this.nFiles = nStorages;
    	this.nFragments = nFragments;
    	this.nNeededFragments = nNeeded;
    	
    	this.accessBrokerList = new Vector<AccessBrokerSimulatorRemote>();
        this.nRegisteredMachines = new AtomicInteger(0);
        this.isSimulationFinished = isSimulationFinished;
        this.isSimulationReadyForStarting = new AtomicBoolean(false);
        this.logger = Logger.getLogger("oppstoreSim.OppStoreSimulator");
        
        try { 
            int bootstrapPort = 9001;        
            bootstrapNodeAddress = new InetSocketAddress(InetAddress.getLocalHost(), bootstrapPort);
        } catch (UnknownHostException e1) { e1.printStackTrace(); }
    }
    
    /** 
     * Sends storage request list to remote clients
     */
    public void simulateFileStorage() {
                
        nStorageResponses = new AtomicInteger(0);
        
        int baseFileSize = 1; // 100k
        
        logger.info("simulateFileStorage -> Starting file storage. nAccessBrokers=" + accessBrokerList.size() +
        		" nFiles=" + nFiles + " nFragments=" + nFragments + " nNeededFragments=" + nNeededFragments + " baseFileSize=" + baseFileSize);
        System.out.println("simulateFileStorage -> Starting file storage.");
        
        for (AccessBrokerSimulatorRemote accessBroker : accessBrokerList) {
            try {
                accessBroker.addFiles(nFiles, nFragments, nNeededFragments, baseFileSize);
                accessBroker.storeFiles();
                
            } catch (RemoteException e) {
                logger.error("simulateFileStorage -> Remote Exception: could not reach AccessBrokerSimulator.");
                logger.error("Cause: " + e.getMessage());
            }            
        }
    }

    /** 
     * Sends data retrieval request list to remote clients
     */
    public void simulateFileRetrieval() {

        logger.info("simulateFileRetrieval -> Starting file retrieval.");
        System.out.println("simulateFileRetrieval -> Starting file retrieval.");
        
        nRetrievalResponses = new AtomicInteger(0);
        
        for (AccessBrokerSimulatorRemote accessBroker : accessBrokerList) {
            try {
                accessBroker.retrieveStoredFiles();
            } catch (RemoteException e) {
                logger.error("simulateDataRetrieval -> could not reach AccessBrokerSimulator.");
                logger.error("Cause: " + e.getMessage());
            }
        }
    }

    // ---------------------------------------------------------------------------------------------
    /**
     * Wait for all remote nodes to finish registering and starts the simulation. 
     */
    public void setRemoteMachineStarted(AccessBrokerSimulatorRemote[] accessBrokerRemoteList) {

    	logger.info("Remote machine started succesfuly!");
    	System.out.println("Remote machine started succesfuly!");
    	
        for ( AccessBrokerSimulatorRemote accessBroker : accessBrokerRemoteList )
            this.accessBrokerList.add(accessBroker);        

        int currentRegisteredMachines = nRegisteredMachines.incrementAndGet();
        if (currentRegisteredMachines > nRemoteMachines) {
            System.out.println("PANIC: OpportuneStoreSimulator -> Too many registered machines!");
            System.exit(-1);
        }
        if (currentRegisteredMachines == nRemoteMachines) {
            System.out.println("All machines registered. Simulation is ready...");
        	this.isSimulationReadyForStarting.set(true);
        }
        
    }
    
    public void launchSimulation() {
    	
        while (this.isSimulationReadyForStarting.get() == false) {
            try { Thread.sleep(1000); } 
            catch (InterruptedException e) { }
        }
        
    	this.simulateFileStorage();
    }

    public void setStorageRequestCompleted(AccessBrokerSimulatorRemote accessBrokerRemote) {
        
        int nResponses = nStorageResponses.incrementAndGet();
        System.out.println("number of storage responses: " + nResponses);
        
        if (nResponses > accessBrokerList.size()) {
            System.out.println("PANIC: OpportuneStoreSimulator -> Too many storage responses!");
            System.exit(-1);
        }
        
        if (nResponses == accessBrokerList.size()) {
            System.out.println("Finished all storage requests.");
            this.simulateFileRetrieval();
        }        
    }

    public void setRetrievalRequestCompleted(AccessBrokerSimulatorRemote accessBrokerRemote) {

        int nResponses = nRetrievalResponses.incrementAndGet();
        System.out.println("number of retrieval responses: " + nResponses);
        
        if (nResponses > accessBrokerList.size()) {
            System.out.println("PANIC: OpportuneStoreSimulator -> Too many retrieval responses!");
            System.exit(-1);
        }
        if (nResponses == accessBrokerList.size()) {
            System.out.println("Finished retrieval request.");
            this.isSimulationFinished.set(true);
            //this.simulateFailures();
        }
    }

    public CdrmBootstrapInformation getBootstrapNodeAddress() {
        
        CdrmBootstrapInformation bootInfo = new CdrmBootstrapInformation();
        //bootInfo.nextCdrmNumber = nextCdrmNumber.getAndAdd(numberOfCdrms);
        bootInfo.bootstrapNodeAddress = bootstrapNodeAddress;
        
        return bootInfo;
    }
    
}
